///////////////////////////////////////////
//@Temperature controller program for PIC-P40 with PIC16F877A
//     by MYCOMKITS.COM
//@ SC1602H is used as LCD
//  The upper 4 bits of B port for data, B0/1/2 for control 
//  timer 1(16bits) for measure temperature
//  timer2(8bits) for PWM
//  PID control
//  RE with timer 0 at every 1ms and auto inc/dec
//  ADC works from -80C to +167C CONFIRMED 20090326
//  NTC Thermister with LOG calicuration using coversion table with 100uA, 10K
//      CONFIRMED 20090329 This works well
//  READ from EEPROM and write to EEPROM
//////////////////////////////////////////
#include <16f877A.h>
#device ADC=10	//this is reqiored
#fuses	HS, NOWDT, PUT, NOPROTECT,NOLVP
#fuses	NOCPD, NOWRT, NODEBUG
#use	fast_io(a)
#use	fast_io(b)
#use	fast_io(c)
#use	fast_io(d)
#use	delay(CLOCK=20000000)
#define	Bmode 0			//port B initial mode all output
#define Amode 0xFF			//port A initial mode all input
#define Cmode 0			//port C initial mode all output
#define Dmode 0xCC			//port D initial mode iiooiioo
//#define shadowB 0x00			//shadow register for portB
#byte	db = 6		//6 is port B address 

//////// Port define and link LCD library 
#define rw PIN_B2		//read/write
#define rs PIN_B1		//chip select
#define stb PIN_B0		//strobe
#include <lcd_lib_nbc2.c>	//non bit manupiration command , because of read modify write trouble
// This library is based on Mr.Gokan's LCD library. Cortesy of Mr.Tetsuya Gokan 

/// port difine for hot and cool LED
#define cool_led PIN_D5		//cool LED
#define hot_led PIN_D4		//hot LED

// port define for output and function
#define output_status PIN_C5		//output status
#define output_switch PIN_D7		//output switch active low
#define function_switch PIN_D6		//output switch active low

//////// Port difine for H driver
#define cool_gate PIN_D0	//active H, when cooling
#define hot_gate PIN_D1		//active H, when hotting

#define EEPROM_p_gain_ADDRESS 0
#define EEPROM_i_gain_ADDRESS 0x08
#define EEPROM_d_gain_ADDRESS 0x10

///// ϐ`
int Prev = 1; //initialize
int Now =10; //initialize
int Prev_o_sw = 1;	//previous status for output switch
int Prev_f_sw = 1;	//previous status for function switch
int function = 0;	// 0=ttemp, 1=sensor, 2=B const, 3=p_gain, 4=i, 5=d, 6=disp duty
int function1 = 1;	//
int i;
int rotaryencoder=0x0C;
short int Prev_re1=1;
static int  count_re1 = 250;	//pulse width initialize
float re_inc;

float  ttemp = 25;         //tearget TE temperature
float  ttemp1;  
float  atemp = 25;         //actual TE temperature
float  ref_temp;	//LM35 reffference value
float sensor_max_temp;
float sensor_min_temp;

float integral = 0;
int p_gain = 50;
int p_gain1 = 50;
int i_gain = 35;
int i_gain1 = 35;
int d_gain = 20;
int d_gain1 = 20;
float e0=0;   //̕΍
float e1=0;   //O̕΍
float e2=0;   //OX̕΍
signed long duty0=0;   //̐

signed long p_out;   //Pp[^
signed long i_out;   //Ip[^
signed long d_out;   //Dp[^

float b_const = 3840;
float b_const1 = 3840;

short int sensor_mode=0;	//1=LM35, 0=NTC, initial sensor type NTC
short int sensor_mode1=0;	//1=LM35, 0=NTC, initial sensor type NTC

short int disp_duty=1;	//1 shows Yes and Duty, initial Duty display mode
short int disp_duty1=1;	//

float Rt;
float log_data2;

// LOG calclation table
const float r_value[17] = {109121.5436, 63185.33369, 37994.24658, 23635.92567, 15161.78685, 10000, 6764.289855, 4682.171292,
	3309.912751, 2385.430628, 1749.917183, 1304.847807, 987.7513058, 758.2094321, 589.5776639, 463.9827971, 369.239262};
const float log_value[17] = {11.60021762, 11.05382749, 10.54519002, 10.07052311, 9.626533518, 9.210340372, 8.819412562,
	8.451517233, 8.104677109, 7.777134944, 7.467323742, 7.17384169, 6.895430951, 6.630959643, 6.379406457, 6.139847476,
	5.911444841};
	
short int sensor_error =0;	// 1 shows error	
int m;
float r_value_temp = 12000;

//void write_float_to_eeprom( float value, unsigned int address );
//void read_float_from_eeprom( float *value, unsigned int address );

// supply LOG data by table
#separate 
float log_calc(float rt_temp) {
	int m=0;
	int n=0;
	float log_data;

	while(rt_temp < r_value[m]) {
		m++;
		n=m;
	}
	log_data = ((rt_temp-r_value[m])/(r_value[m-1]-r_value[m]))*(log_value[m-1]-log_value[m])+log_value[m];	

	return log_data;
}	


///////////////////////////////////////////////
// 
// Display several message on LCD
// with some interval.
// Constant Message is send by lcd_data()
/////////////////////////////////////////////// 

///// ^C}1荞ݏ
// timer1 reads and measures actual temperature at every 1 sec.
//
#int_timer1
void isr1(void)	// 13 bit AD converter , 10 + 3 bits, average of 8 tests
{
	int i;
	set_timer1(0x0001);             // ^C}1Đݒ
	if(sensor_mode) {		//1 shows LM35
		atemp=0;
		set_adc_channel(0);
		for(i = 0;i < 16;++i)	//14bit = AD 10bits +4 bits
		{
			delay_us(50);
			atemp = atemp + read_adc();
		}	
			atemp = (atemp/16)*(sensor_max_temp-sensor_min_temp)/1024+sensor_min_temp; 
		if((atemp>160)||atemp<-60) sensor_error = 1;
		else {
			sensor_error = 0;
		}
	}
	else {	//sensor mode 0 shows NTC
		 set_adc_channel(2);	
		 for(i = 0;i < 16;++i)	//14bit = AD 10bits +4 bits
		{
			delay_us(50);
			atemp = atemp + read_adc();		
		}	
		 atemp = atemp/16;
		 Rt=atemp*48.81815;	//when Ri=100uA
		 if((Rt>109000)||Rt<360) sensor_error = 1;
		else {
			sensor_error = 0;
		 	//log_calc(Rt); //supply LOG data from table
		 	log_data2=log_calc(Rt);
		 	atemp=(1/((1/b_const)*(log_data2-9.2103403)+0.003354))-273.15;
		 }	
	}	 
//
	lcd_cmd(0xC0);	//second line
	if((Prev==1) | (disp_duty==0)) {
		if(sensor_error == 0) 	{
			printf(lcd_data, "Actual T=%2.2fC      ", atemp);   
			//printf(lcd_data, "log_data=%2.6f       ", log_data2); 
		}
		else printf(lcd_data, "Sensor Error            "); 
		if(Now==0) {
			Prev=0;
			Now=10;
		}	
		else Now=Now-1;
	}
	else {
		printf(lcd_data, "PWM Duty=%3.0f       ", (float)duty0); 

		if(Now==0) {
			Prev=1;
			Now=10;
		}	
		else Now=Now-1;
	}	

}

// timer 2 for CCP controller
//
// timer 0 interupt
// read rotary encoder data at every 1msec

#int_timer0
void isr0(void)	// this timer0 may not work every over flow, because of other interrupt.
{
	set_timer0(0x01);             // ^C}0Đݒ
	rotaryencoder = input_d();	//D2 for RE1, D3 for RE2 
}

//
// system intializing
//
//
#separate
void initializing() {
// ADC set them up analog inputs are RA0 RA1 RA2
	setup_adc(ADC_CLOCK_DIV_32);	//every 250nsec
	setup_adc_ports(AN0_AN1_AN2_AN3_AN4);
	set_tris_a(Amode);	//all input
	set_tris_b(Bmode);	//all output for LCD
 	set_tris_c(Cmode);	//all output 
	set_tris_d(Dmode);	//in and out
// motor current zero
	output_C(0xFF);
	output_low(hot_gate);
	output_low(cool_gate);
// LCD test
	lcd_init();				//initialize LCD
	lcd_clear();		//clear display
	lcd_data("MYCOMKITS.COM");
	lcd_cmd(0xC0);	//second line
	lcd_data("internal testing");
	delay_ms(500);	
//	
// all LED test
	output_low(cool_led);
	output_low(hot_led);
	output_high(cool_led);	//cool led
	delay_ms(500);
	output_low(cool_led);
	output_high(hot_led);	//hot led
	delay_ms(500);
	output_low(hot_led);
	output_low(PIN_C5);	//out led
	delay_ms(500);
	output_high(PIN_C5);
	output_low(PIN_C6);	//test led
	delay_ms(500);
	output_high(PIN_C6);
//	
//	
//
	lcd_clear();		//clear display
	ttemp=25.00;
	printf(lcd_data, "Target T=%2.1fC     ", ttemp); 	//
	lcd_cmd(0xC0);	//second line
	atemp=30.00;
	printf(lcd_data, "Actual T=%2.2fC     ", atemp);       //
	Prev=(input_D() & 0x0C);	//RE intialize
	lcd_cmd(0x80);	//1st line
//		
//	PWM䏉ݒ by timer2
//
	setup_ccp1(ccp_pwm);
	setup_ccp2(ccp_pwm);
	setup_timer_2(T2_DIV_BY_1,0xFF,1);
	output_C(0xFF);
	set_pwm2_duty(0xFF);	//length of high, but this is active Low
	set_pwm1_duty(0xFF);
	output_low(cool_gate);
	output_low(hot_gate);
//	
// PID value intialize
//
//// PIDp[^̓ǂݍ  EEPROM //	

	p_gain = read_eeprom(EEPROM_p_gain_ADDRESS);
	i_gain = read_eeprom(EEPROM_i_gain_ADDRESS);
	d_gain = read_eeprom(EEPROM_d_gain_ADDRESS);
// 񂪖ݒ̏ꍇ //
	if((p_gain <= 0) | (p_gain > 100)) {
	p_gain = 50;
	i_gain = 35;
	d_gain = 20;
	write_eeprom(EEPROM_p_gain_ADDRESS, p_gain);
	write_eeprom(EEPROM_i_gain_ADDRESS, i_gain);
	write_eeprom(EEPROM_d_gain_ADDRESS, d_gain);
	}

// read LM35 refference value from A1
	set_adc_channel(1);
	delay_us(50);
	ref_temp = read_adc();
	delay_us(50);
	ref_temp = ref_temp + read_adc();
	ref_temp = ref_temp/2;
//
	sensor_min_temp = 100*(-(ref_temp)*5/1024);	//actual sensor min temp at degree C
	sensor_max_temp = 250-ref_temp*500/1024;	//actual sensor max temp at degree C
//	
// ^C}1ݒ for testing actual temerature by ADC
//
	setup_timer_1(T1_INTERNAL|T1_DIV_BY_8);
	set_timer1(0x0001);              // ^C}1ݒ almost 1 sec
	enable_interrupts(INT_TIMER1);   // ^C}1荞݋
//
// ^C}0ݒ for RotaryEncoder
//	
	setup_timer_0(RTCC_INTERNAL|RTCC_DIV_4);	// one clock = 0.5u = 0.125 * 4
	set_timer0(0x01);              // ^C}0ݒ almost 0.5m sec, when DIV=4
	enable_interrupts(INT_TIMER0);   // ^C}0荞݋
	enable_interrupts(GLOBAL);       // O[o荞݋	
	
} 	
//	
// end of Initialization
//
// PID value calculation and output
//
#separate 
void pid_calc() {	
      e2 = e1;
      e1 = e0;
      e0 = ttemp - atemp;
      integral += e0;
      //
      // ěvZ
      //
      // Pp[^ //
      p_out = p_gain*e0;	//p_gain
      //
      // Ip[^ //
      i_out = i_gain*integral;	//i_gain * summing of positive or negative area
      //
      // Dp[^ //
      d_out =d_gain*(e1-e0);	//d_gain * (atemp-atemp(n-1))
      //
      // set duty
      duty0 = p_out+i_out+d_out;
      if(duty0 > 255) duty0=255; 
      if(duty0 < -255) duty0=-255; 
      //
      // drive peltie cool or hot//
      //
      if(!(input(output_status))) {	//test output status, L shows ON 
      	if(duty0<0) {	//cooling
	      	if(hot_gate == 1) integral = 0;
			output_low(hot_gate);	//active H
			output_low(hot_led);	//active H LED
			set_pwm2_duty(0xFF);	//length of high, but this is active Low
			output_high(cool_gate);	//active H
			output_high(cool_led);	//active H
			set_pwm1_duty(255+duty0);	//length of high, but this is active Low
		}
		else {
			if(cool_gate == 1) integral = 0;
			output_low(cool_gate);	//active H
			output_low(cool_led);	//active H LED
			set_pwm1_duty(0xFF);	//length of high, but this is active Low
			output_high(hot_gate);	//active H
			output_high(hot_led);	//active H LED
			set_pwm2_duty(255-duty0);	//length of high, but this is active Low
      	}
      }	
	else {	
		output_low(hot_gate);
		set_pwm2_duty(0xFF);	//length of high, but this is active Low	
		output_low(cool_gate);
		set_pwm1_duty(0xFF);	//length of high, but this is active Low
		output_low(hot_led);	//active H LED
		output_low(cool_led);	//active H LED
	}			
}	
// rotary encoder 
// sampling method, every 1msec with shadow register
// with integration method
//
#separate 
float r_encoder(float base_v, signed long min_v, signed long max_v, float step_s, float step_l) {
	int re_v;
	//output_toggle(PIN_C6);	//for test
	re_v = rotaryencoder & 0x0C;	//D2 for RE1, D3 for RE2
		if(!(re_v & 0x04)) {
			if(Prev_re1) {
				if(re_v & 0x08) re_inc = step_s;	//when RE2=H, it is CW
				else re_inc = -(step_s);		
				base_v = base_v + re_inc;
				if(base_v > max_v) base_v = max_v;
				if(base_v < min_v) base_v = min_v;						
				Prev_re1 = 0;	
				count_re1 = 0;
			}
			else {
				count_re1 = count_re1 + 1;	//count length of low
				if(count_re1>=250) {	// wait around 1 sec up to auto inc/dec
					count_re1=250;
					if(re_inc >= 0) re_inc = step_l;	//when RE2=H, it is CW
					else re_inc = -(step_l);
					base_v = base_v + re_inc;
					if(base_v > max_v) base_v = max_v;
					if(base_v < min_v) base_v = min_v;	
					//output_toggle(PIN_C6);	//for test
				}
			}		
		}		
		else {
			Prev_re1 = 1;
			count_re1 = 0;
		}
	return base_v;	
}		
// rotary encoder test end//
// main program
//

void main(){
//	
// initialzing
	initializing();
//	
	while(1){			//endless loop
//
//Out SW test
//
		if(!input(output_switch)) {	//output_switch is PIN_D7 active low
			delay_ms(50);	//check it again for chattaling
			if(!input(output_switch))	 {
				if(Prev_o_sw) output_toggle(output_status);	//output on led toggle, PIN_C5 shows output status
				Prev_o_sw = 0;
			}
		}
		else
		Prev_o_sw = 1;
//		
//
// Out SW test end
//	
// function switch test
//
		if(!input(function_switch)) {	//function_switch is PIN_D6 active low
			delay_ms(50);	//check it again for chattaling
			if(!input(function_switch))	 {
				if(Prev_f_sw==1) {
					function ++;	//increment function mode
					if(function>=7) function =0;
					Prev_f_sw = 0;
					function1=1;
				}
			}
		}
		else Prev_f_sw = 1;
//			

		switch(function) {
			case 0: ttemp1=ttemp;	//1st line 
					ttemp = r_encoder(ttemp, -50, 150, 0.1, 1);
					if(!(ttemp==ttemp1) | function1) {	//if no change, do not print. this is solution
						lcd_cmd(0x80);
						printf(lcd_data, "Target T=%2.1fC     ", ttemp);  //10digit+
						function1=0;
					}
					break;
			case 1: 	sensor_mode1=sensor_mode;	//mode 1 shows LM35

					sensor_mode=r_encoder(sensor_mode, -255, 255, 1, 1); 
					sensor_mode %= sensor_mode/2;
					if(!(sensor_mode==sensor_mode1) | function1) {
						lcd_cmd(0x80);	//1st line
						printf(lcd_data, "NTCorLM35? "); 	//11 cahr
						if(sensor_mode==1) printf(lcd_data, "LM35       "); 	//mode 1 shows LM35
						else printf(lcd_data, "NTC       "); 	//
						function1=0;
					}	
					break;	
			case 2: if(sensor_mode==0) {	//0 shows NTC
						b_const1 = b_const;
						b_const = r_encoder(b_const, 2000, 5000, 1, 10);
						if(!(b_const==b_const1) | function1) {	//if no change, do not print. this is solution
							lcd_cmd(0x80);	//1st line
							printf(lcd_data, "B const =%4.0f          ", b_const); 	//	
							function1=0;
						}	
					}
					else {	//1 shows LM35, 
						function ++;
					}
					break;
			case 3:	p_gain1=p_gain;	//1st line 
					p_gain = r_encoder(p_gain, 10, 100, 1, 5);
					if(!(p_gain==p_gain1) | function1) {	//if no change, do not print. this is solution
						lcd_cmd(0x80);
						printf(lcd_data, "p_gain?=%3.0f          ",(float)p_gain); 	//
						function1=0;
					}					
					break;	
			case 4:	i_gain1=i_gain;	//1st line 
					i_gain = r_encoder(i_gain, 5, 70, 1, 5);
					if(!(i_gain==i_gain1) | function1) {	//if no change, do not print. this is solution
						lcd_cmd(0x80);
						printf(lcd_data, "i_gain?=%3.0f          ",(float)i_gain); 	//
						function1=0;
					}					
					break;		
			case 5:	d_gain1=d_gain;	//1st line 
					d_gain = r_encoder(d_gain, 1, 40, 1, 2);
					if(!(d_gain==d_gain1) | function1) {	//if no change, do not print. this is solution
						lcd_cmd(0x80);
						printf(lcd_data, "d_gain?=%3.0f          ",(float)d_gain); 	//
						function1=0;
					}					
					break;		
			case 6:	disp_duty1=disp_duty;	//1st line 
					disp_duty = r_encoder(disp_duty, -255, 255, 1, 1);
					disp_duty %= disp_duty/2;
					if(!(disp_duty==disp_duty1) | function1) {	//if no change, do not print. this is solution
						lcd_cmd(0x80);
						printf(lcd_data, "Display Duty?"); 	//13 cahr
						if(disp_duty==1) printf(lcd_data, "YES "); 	//mode 1 shows duty
						else printf(lcd_data, "NO  "); 	//
						function1=0;
					}					
					break;							
		}	
		
//

//		
//pid calculation and output
		pid_calc();	
		
	}
} 
